# coding=utf-8
"""Defines the network test suite."""
import torch
from torch.utils import data

import math
import numpy as np
from sklearn import metrics as skl_metrics

from .utils import round_list, _get_probs, _filter_matched_subj
from ..plotters.visualization import display_confusion_matrix
from collections import defaultdict

import copy
import inspect

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


# noinspection PyProtectedMember
class Metrics(object):
    """
    A class to calculate all metrics for a BaseNetwork.

    Responsible for the test suite.
    """

    def __init__(self):
        """Initialize the metrics class for a BaseNetwork."""

    # TODO: consider making class_converted default True
    @staticmethod
    def get_score(targets, predictions, metrics='accuracy', average=None,
                  class_converted=False):
        """
        Calculate the provided metrics given some targets and predictions.

        Parameters:
            targets: numpy.ndarray of integers
                The target values
            predictions: numpy.ndarray of integers or numpy.ndarray of floats
                The predicted values
            metrics: list of strings
                The strings definition the "get_"... functions to call
            average: string,
                [None, ‘binary’ (default), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on the
                data.See scikit learn
            class_converted: binary. default False
                True: If raw_predictions have already been converted using
                transform_outputs
                False: If raw_predictions are used

        Returns:
            metrics: dict
                Values returned by the metric functions.

        """
        functions_without_average_parameter = ["get_accuracy", "get_mse"]

        if isinstance(targets, torch.Tensor):
            targets = targets.cpu().detach().numpy()

        if not class_converted:
            predictions = Metrics.transform_outputs(predictions)

        if isinstance(metrics, str):
            metrics = [metrics]

        results_dict = {}

        for metric in metrics:
            method_name = "get_" + metric
            method = getattr(Metrics, method_name)
            try:
                if method_name in set(functions_without_average_parameter):
                    metric_results = method(targets, predictions)
                else:
                    metric_results = method(targets, predictions, average)

            except (AttributeError, TypeError):
                logger.warning(
                    "Metric {} does not exist via function {}().".format(
                        metric, method_name))

            results_dict[metric] = metric_results

        return results_dict

    @staticmethod
    def transform_outputs(in_matrix):
        """
        Reformat output matrix.

        If one-hot, truth matrix to be the classes in a 1D array.

        Note: This does not handle multiple class prediction.

        Parameters:
            in_matrix : numpy.ndarray or torch.Tensor
                One-hot matrix of shape [batch, num_classes].

        Returns:
            class_list : numpy.ndarray of floats
                1D class array.

        """
        if isinstance(in_matrix, torch.Tensor):
            in_matrix = in_matrix.cpu().detach().numpy()


        # For one-hot encoded entries
        if in_matrix.shape[1] > 1:
            return np.argmax(in_matrix, axis=1).astype(np.float32)
        # For single value entries
        elif in_matrix.shape[1] == 1:
            return in_matrix
        else:
            return None

    @staticmethod
    def get_confusion_matrix_values(targets, predictions):
        """
        Calculate the tp, tn, fp, fn values given targets and predictions.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.

        Returns:
            tp, tn, fp, fn:  array of np.float32 with classes in sorted order
                The true positive/negative, false positive/negative values.

        """
        # credit: Robert Fratila
        confusion_matrix = skl_metrics.confusion_matrix(targets, predictions)
        tp = np.diagonal(confusion_matrix).astype('float32')
        tn = (np.array(
            [np.sum(confusion_matrix)] *
            confusion_matrix.shape[0]) -
            confusion_matrix.sum(axis=0) -
            confusion_matrix.sum(axis=1) + tp).astype('float32')
        # sum each column and remove diagonal
        fp = (confusion_matrix.sum(axis=0) - tp).astype('float32')
        # sum each row and remove diagonal
        fn = (confusion_matrix.sum(axis=1) - tp).astype('float32')

        return tp, tn, fp, fn

    @staticmethod
    def _check_average_parameter(targets, predictions, average):
        """
        Check to see if average parameter is suitable for the data.

        Parameters:
            targets: numpy.ndarray of integers
                The target values
            predictions: numpy.ndarray of integers
                The predicted values
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See scikit learn.

        Raises:
            ValueError if `average` value invalid for type of predictions.

        Returns:
            boolean

        """
        if not average:
            return True

        if np.unique(predictions).size <= 2 and np.unique(targets).size <= 2:
            if "binary" not in average:
                raise ValueError(
                    "You must provide binary as the average \
                    function if binary data"
                )

        if np.unique(predictions).size > 2 or np.unique(targets).size > 2:
            if "binary" in average:
                raise ValueError(
                    "You cannot provide binary as the average \
                    function if non binary data"
                )

        return True

    @staticmethod
    def get_sensitivity(targets, predictions, average=None, pos_label=1):
        """
        Calculate the sensitivity.

        Also referred to as recall, or the true positive rate.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.
            pos_label: int
                0 or 1. Only used in the binary case to indicate which class
                label will be used to provide the output for average=binary.


        Returns:
            sensitivity: np.float32 or array of np.float32
                The sensitivity.

        """
        assert Metrics._check_average_parameter(targets, predictions, average)
        sensitivity = skl_metrics.recall_score(targets, predictions,
                                               average=average,
                                               pos_label=pos_label)
        return sensitivity

    @staticmethod
    def get_specificity(targets, predictions, average=None, pos_label=1):
        """
        Calculate the specificity.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.
            pos_label: int
                0 or 1. Only used in the binary case to indicate which class
                label will be used to provide the output for average=binary.

        Returns:
            specificity: np.float32 or array of np.float32
                The specificity.

        """
        _, tn, fp, _ = Metrics.get_confusion_matrix_values(targets,
                                                           predictions)

        specificity = tn / (tn + fp)

        # TODO: implement other options
        if average == "macro":
            specificity = np.average(specificity)
        elif average == "binary":
            specificity = specificity[pos_label]
        elif average:
            raise NotImplementedError

        return specificity

    @staticmethod
    def get_dice(targets, predictions, average=None, pos_label=1):
        """
        Calculate the dice metric.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                Only None, 'binary', and 'macro' currently implemented.
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.
            pos_label: int
                0 or 1. Only used in the binary case to indicate which class
                label will be used to provide the output for average=binary.

        Returns:
            dice: np.float32 or array of np.float32
                The dice metric.

        """
        tp, _, fp, fn = Metrics.get_confusion_matrix_values(targets,
                                                            predictions)
        dice = 2 * tp / (2 * tp + fp + fn)

        # TODO: implement other options
        if average == "macro":
            dice = np.average(dice)
        elif average == "binary":
            dice = dice[pos_label]
        elif average:
            raise NotImplementedError

        return dice

    @staticmethod
    def get_ppv(targets, predictions, average=None, pos_label=1):
        """
        Calculate the positive predictive value.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                Only None, 'binary', and 'macro' currently implemented.
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.
            pos_label: int
                0 or 1. Only used in the binary case to indicate which class
                label will be used to provide the output for average=binary.

        Returns:
            ppv: float32 or array of np.float32
                The positive predictive value.

        """
        assert Metrics._check_average_parameter(targets, predictions,
                                                average=average)
        ppv = skl_metrics.precision_score(targets, predictions,
                                          average=average, pos_label=pos_label)
        return ppv

    @staticmethod
    def get_npv(targets, predictions, average=None, pos_label=1):
        """
        Calculate the negative predictive value.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                Only None, 'binary', and 'macro' currently implemented.
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.
            pos_label: int
                0 or 1. Only used in the binary case to indicate which class
                label will be used to provide the output for average=binary.

        Returns:
            npv: np.float32 or array of np.float32
                The negative predictive value

        """
        _, tn, _, fn = Metrics.get_confusion_matrix_values(targets,
                                                           predictions)

        npv = np.nan_to_num(tn / (tn + fn))

        # TODO: implement other options
        if average == "macro":
            npv = np.average(npv)
        elif average == "binary":
            npv = npv[pos_label]
        elif average:
            raise NotImplementedError

        return npv

    @staticmethod
    def get_accuracy(targets, predictions):
        """
        Calculate the accuracy.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            predictions: numpy.ndarray of integers
                The predicted values.

        Returns:
            accuracy: array of np.float32
                The accuracy.

        """
        accuracy = skl_metrics.accuracy_score(targets, predictions)

        return accuracy

    @staticmethod
    def get_f1(targets, predictions, average=None, pos_label=1):
        """
        Calculate the f1 score.

        Parameters:
            targets: numpy.ndarray of integers
                The target values
            predictions: numpy.ndarray of integers
                The predicted values
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.

        Returns:
            f1: np.float32 or array of np.float32
                The f1 score.

        """
        assert Metrics._check_average_parameter(targets, predictions, average)
        f1 = skl_metrics.f1_score(targets, predictions, average=average,
                                  pos_label=pos_label)

        return f1

    # TODO: what type are the raw predicted values that come out??
    @staticmethod
    def get_auc(targets, raw_predictions, num_classes, average=None, pos_label=1):
        """
        Calculate the AUC. Note: raw_predictions and num_classes are required.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            raw_predictions: numpy.ndarray of floats
                The raw predicted values, not converted to classes.
            average: string
                [None, ‘binary’ (def), ‘micro’, ‘macro’, ‘samples’, ‘weighted’]
                This parameter is required for multiclass/multilabel targets.
                If None, the scores for each class are returned.
                Otherwise, this determines the type of averaging performed on
                the data. See Scikit learn.

        Returns:
            f1: np.float32 or array of np.float32
                The AUC.

        """
        if raw_predictions.ndim == 1:
            raise ValueError("You must provide raw predictions not \
                                    class_converted predictions")

        all_class_auc = []
        for i in range(num_classes):
            if num_classes == 1:
                fpr, tpr, _ = skl_metrics.roc_curve(targets,
                                                    raw_predictions,
                                                    pos_label=1)
            else:
                fpr, tpr, _ = skl_metrics.roc_curve(targets,
                                                    raw_predictions[:, i],
                                                    pos_label=i)
            auc = skl_metrics.auc(fpr, tpr)
            all_class_auc += [auc]
        # TODO: implement other options
        if average == "macro":
            all_class_auc = np.average(all_class_auc)
        elif average == "binary":
            all_class_auc = all_class_auc[pos_label]
        elif average:
            raise NotImplementedError

        return all_class_auc

    @staticmethod
    def get_mse(targets, raw_predictions):
        """
        Calculate the negative MSE.

        Note: raw_predictions and num_classes are required. Negative
        values are returned so the value can be optimized.

        Parameters:
            targets: numpy.ndarray of integers
                The target values.
            raw_predictions: numpy.ndarray of floats
                The raw predicted values, not converted to classes.

        Returns:
              results: float
                The MSE value

        """
        return skl_metrics.mean_squared_error(targets, raw_predictions)

    @staticmethod
    def run_test(network, data_loader, plot=False, save_path=None,
                 pos_label=1, transform_callable=None, **kwargs):
        """
        Will conduct the test suite to determine network strength.

        Calls either _run_test_single_continuous or _run_test_multi
        depending on the number of classes. _run_test_multi
        only works with raw one-hot encoded output values. Note that multi
        class multi outputs are not supported: values will either be converted
        to class values from one-hot encoding or be a single continuous value.
        However, values can be transformed after the forward pass using
        transform_callable.

        Parameters:
            network: nn.Module
                The network
            data_loader : DataLoader
                A DataLoader object to run the test with.
            save_path : string
                Folder to place images in.
            plot: bool
                Determine if graphs should be plotted in real time.
            pos_label: int
                The label that is positive in the binary case for macro
                calculations.
            transform_callable: callable
                A torch function. e.g. torch.round()
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)

        Returns:
            results : dict

        """
        num_classes = network._num_classes

        if num_classes is None or num_classes == 0:
            raise ValueError('There\'s no classification layer')
        elif num_classes == 1:
            results_dict = \
                Metrics._run_test_single_continuous(
                    network,
                    data_loader,
                    transform_callable=transform_callable,
                    **kwargs)
        else:
            results_dict = Metrics._run_test_multi(network, data_loader,
                                                   plot=plot, save_path=save_path,
                                                   pos_label=pos_label)

        return results_dict

    @staticmethod
    def _run_test_single_continuous(network, data_loader,
                                    transform_callable, **kwargs):
        """
        Will conduct the test suite to determine network strength.

        Parameters:
            network: nn.Module
                The network
            data_loader : DataLoader
                A DataLoader object to run the test with.
            transform_callable: callable
                A torch function. e.g. torch.round()

        Returns:
            results : dict

        """
        targets = np.array([v[1] for v in data_loader.dataset])

        predictions = network.forward_pass(
            data_loader=data_loader,
            convert_to_class=False,
            transform_callable=transform_callable
            **kwargs)


        mse = Metrics.get_mse(targets, predictions)

        logger.info('{} test\'s results'.format(network.name))

        logger.info('\nMean Squared Error: {}'.format(mse))

        return {"mse": mse}

    @staticmethod
    def _run_test_multi(network, data_loader, plot=False, save_path=None,
                        pos_label=1, transform_callable=None):
        """
        Will conduct the test suite to determine network strength.

        No transforms will be conducted on the outputs, save where
        necessary for calculating metric values. You can transform values
        after the forward pass by using transform_callable.

        Parameters:
            network: nn.Module
                The network
            data_loader : DataLoader
                A DataLoader object to run the test with.
            save_path : string
                Folder to place images in.
            plot: bool
                Determine if graphs should be plotted in real time.
            transform_callable: callable
                A torch function. e.g. torch.round()

        Returns:
            results : dict

        """
        num_classes = network._num_classes

        if num_classes > 2:
            average = "macro"
        elif num_classes == 2:
            logger.warning("Will report scores only for pos_label, which is \
                           set to {}".format(pos_label))
            average = "binary"

        if plot:
            logger.setLevel(logging.INFO)

        # getting just the y values out of the dataset
        # TODO: store in tensor for continuity?
        targets = np.array([v[1] for v in data_loader.dataset])

        raw_predictions = network.forward_pass(
            data_loader=data_loader,
            transform_callable=transform_callable
        )

        predictions = Metrics.transform_outputs(raw_predictions)

        cm = skl_metrics.confusion_matrix(targets, predictions)
        if plot:
            display_confusion_matrix(cm, save_path=save_path)

        tp, tn, fp, fn = Metrics.get_confusion_matrix_values(targets,
                                                             predictions)

        sensitivity = Metrics.get_sensitivity(targets, predictions)
        sensitivity_macro = Metrics.get_sensitivity(targets, predictions,
                                                    average=average)

        specificity = Metrics.get_specificity(targets, predictions)
        specificity_macro = Metrics.get_specificity(targets, predictions,
                                                    average=average)

        dice = Metrics.get_dice(targets, predictions)
        dice_macro = Metrics.get_dice(targets, predictions, average=average)

        ppv = Metrics.get_ppv(targets, predictions)
        ppv_macro = Metrics.get_ppv(targets, predictions, average=average)

        npv = Metrics.get_npv(targets, predictions)
        npv_macro = Metrics.get_npv(targets, predictions, average=average)

        accuracy = Metrics.get_accuracy(targets, predictions)

        f1 = Metrics.get_f1(targets, predictions)
        f1_macro = Metrics.get_f1(targets, predictions, average=average)

        auc = Metrics.get_auc(targets, raw_predictions, num_classes)
        auc_macro = Metrics.get_auc(targets, raw_predictions, num_classes,
                                    average=average)

        logger.info('{} test\'s results'.format(network.name))

        logger.info('TP: {}'.format(tp))
        logger.info('FP: {}'.format(fp))
        logger.info('TN: {}'.format(tn))
        logger.info('FN: {}'.format(fn))

        logger.info('\nAccuracy: {}'.format(accuracy))

        logger.info('Sensitivity: {}'.format(round_list(sensitivity,
                                                        decimals=3)))
        # accuracy and macro sensitivity are the same when classes are balanced
        logger.info('\tMacro Sensitivity: {:.4f}'.format(sensitivity_macro))

        logger.info('Specificity: {}'.format(round_list(specificity,
                                                        decimals=3)))
        logger.info('\tMacro Specificity: {:.4f}'.format(specificity_macro))

        logger.info('DICE: {}'.format(round_list(dice, decimals=3)))
        logger.info('\tAvg. DICE: {:.4f}'.format(np.average(dice)))

        logger.info('Positive Predictive Value: {}'.format(
            round_list(ppv, decimals=3)))
        logger.info('\tMacro Positive Predictive Value: {:.4f}'.format(
            ppv_macro))

        logger.info('Negative Predictive Value: {}'.format(
            round_list(npv, decimals=3)))
        logger.info('\tMacro Negative Predictive Value: {:.4f}'.format(
            npv_macro))

        logger.info('F1-score: {}'.format(round_list(f1, decimals=3)))
        logger.info('\tMacro f1-score: {:.4f}'.format(f1_macro))

        logger.info('auc: {}'.format(round_list(auc, decimals=3)))
        logger.info('\tMacro auc: {:.4f}'.format(auc_macro))

        return {
            'accuracy': float(accuracy),
            'macro_sensitivity': float(sensitivity_macro),
            'macro_specificity': float(specificity_macro),
            'avg_dice': float(dice_macro),
            'macro_ppv': float(ppv_macro),
            'macro_npv': float(npv_macro),
            'macro_f1': float(f1_macro),
            'macro_auc': float(auc_macro)
        }
    @staticmethod
    def bootfold_p_estimate(network, data_loader, n_samples, k, epochs,
			    index_to_iter, ls_feat_vals, 
			    retain_graph=None, valid_interv=4, plot=False, 
                            save_path=None, transform_outputs=False, transform_callable=None, **kwargs):
        """
        Performs bootfold - estimation to identify whether training model provides statistically significant
        difference in predicting various values for a given feature when predicting outcome.

        Parameters:
            network : BaseNetwork
                Network descendant of BaseNetwork.
            data_loader : torch.utils.data.DataLoader
                The DataLoader object containing the totality of the data to use
                for k-fold cross validation.
            n_samples : int
                number of times to randomly sample w/ replacement the data_loader and perform boot_cv
            k : int
                The number of folds to split the training into.
            epochs : int
                The number of epochs to train the network per fold.
            index_to_iter : string
                Name of feature within data_loader who's values will be iterated to assess difference
            ls_feat_vals : list
                List of values for feature provided in index_to_iter
            retain_graph : {None, boolean}
                Whether retain_graph will be true when .backwards is called.
            valid_interv : int
                Specifies after how many epochs validation should occur.
            plot : boolean
                Whether or not to plot all results in prompt and charts.
            save_path : str
                Where to save all figures and results.
            transform_outputs : boolean
                Not used in the multi-class case.
                If true, transform outputs using metrics.transform_outputs.
                If no transform_callable is provided then the defaults in
                metrics.transform_outputs will be used: class converstion for
                one-hot encoded, and identity for one-dimensional outputs.
                Multiple class multiple outputs are not yet supported.
            transform_callable: callable
                Not used in the multi-class case.
                Used to transform values if transform_outputs is true,
                otherwise defaults in metrics.transform_outputs will be used.
                An example could be np.round
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)

        Returns:
            p_value : float
        """
        ls_imprv_scores = []

        rand_sample = data.RandomSampler(data_loader.dataset, replacement=True)
        data_loader_args = inspect.signature(data.DataLoader.__init__)
        new_params = {}

        for param in data_loader_args.parameters:
            if param == "self":
                continue
            elif param == "shuffle":
                continue
            elif param == "sampler":
                new_params["sampler"] = rand_sample
            elif param == "batch_sampler":
                new_params["batch_sampler"] = None
            else:
                new_params[param] = getattr(data_loader, param)

        dl_sample = data.DataLoader(**new_params)

        for samp in range(n_samples):
            imprv_score = Metrics._boot_cv(network, dl_sample, k, epochs, retain_graph, 
                valid_interv, plot, save_path, index_to_iter, ls_feat_vals)
            ls_imprv_scores.append(imprv_score)

        tot_num_imprv = float(len(ls_imprv_scores))
        score_below_zero = sum(val <= 0.0 for val in ls_imprv_scores)
        # calculate p value based on change of not improving 
        p_val = float(score_below_zero)/float(tot_num_imprv)
        logger.info("Improvement scores: {}".format(', '.join(map(str, ls_imprv_scores))))
        logger.info("P value for bootfold p estimate: %f.", p_val)

    def _boot_cv(network, data_loader, k, epochs, retain_graph, valid_interv, 
                save_path, plot, index_to_iter, ls_feat_vals):
        """
        Perform a custom cross validation for bootstrapped p estimation.

        Parameters:
            network : BaseNetwork
                Network descendant of BaseNetwork.
            data_loader : torch.utils.data.DataLoader
                The DataLoader object containing the totality of the data to use
                for k-fold cross validation.
            k : int
                The number of folds to split the training into.
            epochs : int
                The number of epochs to train the network per fold.
            retain_graph : {None, boolean}
                Whether retain_graph will be true when .backwards is called.
            valid_interv : int
                Specifies after how many epochs validation should occur.
            plot : boolean
                Whether or not to plot all results in prompt and charts.
	    index_to_iter : string
                Index of feature within data_loader who's values will be iterated to assess difference
            ls_feat_vals : list
                List of values for feature provided in index_to_iter

	Returns: 
	    improvement_score : float
        """
        dct_improvementScores = defaultdict(list)
        dct_filteredSubj = defaultdict(list)
        ls_filteredProbs = []
	
        fold_len = math.floor(len(data_loader.dataset) / k)
        rem = len(data_loader.dataset) % k
        fold_seq = []

        for _ in range(k-1):
            fold_seq.append(fold_len)
        if rem == 0:
            fold_seq.append(fold_len)
        else:
            fold_seq.append(fold_len+rem)  # last one is the longest if unequal

        dataset_splits = data.random_split(data_loader.dataset,
                                           fold_seq)

        batch_size = data_loader.batch_size
        shuffle = isinstance(data_loader.sampler, data.sampler.RandomSampler)

        try:
            for fold in range(k):

                # TODO: this may break on different devices?? test.
                # TODO: Re-initialize instead of deepcopy?
                cross_val_network = copy.deepcopy(network)

                # TODO: properly pass params

                # Generate fold training set.
                train_dataset = data.ConcatDataset(
                    dataset_splits[:fold] + dataset_splits[fold+1:])
                # Generate fold validation set.
                val_dataset = dataset_splits[fold]
                # Generate fold training data loader object.
                train_loader = data.DataLoader(
                    train_dataset, batch_size=batch_size, shuffle=shuffle)
                # Generate fold validation data loader object.
                val_loader = data.DataLoader(
                    val_dataset, batch_size=batch_size)
                # Train network on fold training data loader.
                cross_val_network.fit(
                    train_loader, val_loader, epochs,
                    retain_graph=retain_graph,
                    valid_interv=valid_interv, plot=plot, save_path=save_path)
                dct_scores = _get_probs(network, val_loader, index_to_iter, ls_feat_vals)
                dct_filtered = _filter_matched_subj(dct_scores, val_loader, index_to_iter)
                for ind in dct_filtered:
                    dct_filteredSubj[ind].append(dct_filtered[ind])
                    ls_filteredProbs.append(dct_filtered[ind])
            V_p = np.array(ls_filteredProbs).mean()
            ls_pos_rate = [subj for subj in data_loader.dataset if subj[1].item() == 1]
            V_t = float(len(ls_pos_rate)/len(data_loader.dataset)) * 100
            improvement_score = float(V_p/V_t)
            if np.isnan(improvement_score):
                improvement_score = 0.0
            return improvement_score
        except KeyboardInterrupt:
            logger.info(
                "\n\n***KeyboardInterrupt: Cross validate stopped \
                prematurely.***\n\n")
	
    @staticmethod
    def cross_validate(network, data_loader, k, epochs,
                       average_results=True, retain_graph=None,
                       valid_interv=4, plot=False, save_path=None,
                       transform_callable=None,
                       **kwargs):
        """
        Perform k-fold cross validation given a Network and DataLoader object.

        Parameters:
            network : BaseNetwork
                Network descendant of BaseNetwork.
            data_loader : torch.utils.data.DataLoader
                The DataLoader object containing the totality of the data to use
                for k-fold cross validation.
            k : int
                The number of folds to split the training into.
            epochs : int
                The number of epochs to train the network per fold.
            average_results : boolean
                Whether or not to return results from all folds or just an average.
            retain_graph : {None, boolean}
                Whether retain_graph will be true when .backwards is called.
            valid_interv : int
                Specifies after how many epochs validation should occur.
            plot : boolean
                Whether or not to plot all results in prompt and charts.
            save_path : str
                Where to save all figures and results.
            transform_callable: callable
                A torch function. e.g. torch.round()
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)


        Returns:
            results : dict
                If average_results is on, return dict of floats.
                If average_results is off, return dict of float lists.

        """
        all_results = defaultdict(lambda: [])

        # TODO: this whole section is really clunky
        # Getting the fold sequence.
        fold_len = math.floor(len(data_loader.dataset) / k)
        rem = len(data_loader.dataset) % k
        fold_seq = []

        for _ in range(k-1):
            fold_seq.append(fold_len)
        if rem == 0:
            fold_seq.append(fold_len)
        else:
            fold_seq.append(fold_len+rem)  # last one is the longest if unequal

        dataset_splits = data.random_split(data_loader.dataset,
                                           fold_seq)

        batch_size = data_loader.batch_size

        # #TODO: improve the copying of parameters
        # Set to true if RandomSampler exists.
        shuffle = isinstance(data_loader.sampler, data.sampler.RandomSampler)

        try:
            for fold in range(k):

                # TODO: this may break on different devices?? test.
                # TODO: Re-initialize instead of deepcopy?
                cross_val_network = copy.deepcopy(network)

                # TODO: properly pass params

                # Generate fold training set.
                train_dataset = data.ConcatDataset(
                    dataset_splits[:fold] + dataset_splits[fold+1:])
                # Generate fold validation set.
                val_dataset = dataset_splits[fold]
                # Generate fold training data loader object.
                train_loader = data.DataLoader(
                    train_dataset, batch_size=batch_size, shuffle=shuffle)
                # Generate fold validation data loader object.
                val_loader = data.DataLoader(
                    val_dataset, batch_size=batch_size)
                # Train network on fold training data loader.
                cross_val_network.fit(
                    train_loader, val_loader, epochs,
                    retain_graph=retain_graph,
                    valid_interv=valid_interv, plot=plot, save_path=save_path)
                # Validate network performance on validation data loader.
                results = Metrics.run_test(
                    cross_val_network, val_loader,
                    save_path=save_path, plot=plot,
                    transform_callable=transform_callable,
                    **kwargs)

                logger.info(results)
                for m in results:
                    all_results[m].append(results[m])

        # TODO: we could show something better here like calculate
        # all the results so far
        except KeyboardInterrupt:
            logger.info(
                "\n\n***KeyboardInterrupt: Cross validate stopped \
                prematurely.***\n\n")

        if average_results:
            averaged_all_results = {}
            for m in all_results:
                averaged_all_results[m] = np.mean(all_results[m])
            logger.info(averaged_all_results)
            return averaged_all_results

        else:
            logger.info(all_results)
            return all_results




"""Test all ConvNet capabilities."""
import pytest
import numpy as np
import copy
import logging
import os
import shutil

import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Subset, TensorDataset

import vulcanai
from vulcanai.models import BaseNetwork
from vulcanai.models.cnn import ConvNet, ConvNetConfig
from vulcanai.models.utils import master_device_setter

logger = logging.getLogger(__name__)


class TestConvNet:
    """Define ConvNet test class."""

    @pytest.fixture
    def multi_input_cnn_train_loader(self, multi_input_cnn_data):
        """Synthetic test data pytorch dataloader object."""
        test_train = Subset(multi_input_cnn_data,
                            range(len(multi_input_cnn_data)//2))
        return DataLoader(test_train, batch_size=2)

    @pytest.fixture
    def multi_input_cnn_test_loader(self, multi_input_cnn_data):
        """Synthetic train data pytorch dataloader object."""
        test_val = Subset(multi_input_cnn_data,
                          range(len(multi_input_cnn_data)//2,
                                len(multi_input_cnn_data)))
        return DataLoader(test_val, batch_size=2)

    def test_init(self, conv1D_net):
        """Initialization Test of a ConvNet object."""
        assert isinstance(conv1D_net, BaseNetwork)
        assert isinstance(conv1D_net, nn.Module)
        assert hasattr(conv1D_net, 'network')
        assert hasattr(conv1D_net, 'in_dim')
        assert hasattr(conv1D_net, 'record')
        assert hasattr(conv1D_net, 'device')

        assert conv1D_net._name is not None
        assert isinstance(conv1D_net._config, ConvNetConfig)

        assert conv1D_net.input_networks is None
        assert conv1D_net.epoch == 0
        assert conv1D_net.optim is None
        assert conv1D_net.criterion is None

        assert not hasattr(conv1D_net, 'metrics')

    def test_function_multi_input(self, conv1D_net, multi_input_cnn):
        """Test functions wrt multi_input_cnn."""
        tensor_size = [10, 1, 28, 28]
        assert isinstance(multi_input_cnn.input_networks, nn.ModuleDict)
        assert len(list(multi_input_cnn.input_networks)) == 3
        assert all(multi_input_cnn._get_max_incoming_spatial_dims() ==
                   (8, 8, 8))
        # TODO: make more elegant
        assert multi_input_cnn._merge_input_network_outputs([
                torch.rand(size=tensor_size),
                torch.rand(size=tensor_size + [28]),
                torch.rand(size=[10] +
                           list(multi_input_cnn.input_networks
                           ['multi_input_dnn'].out_dim))
        ]).shape == (10, 3, 8, 8, 8)
        test_net = copy.deepcopy(multi_input_cnn)
        test_net._add_input_network(conv1D_net)
        assert len(list(test_net.input_networks)) == 4
        assert 'conv1D_net' in test_net.input_networks
        assert isinstance(test_net.input_networks['conv1D_net'], ConvNet)

    def test_forward(self, conv1D_net):
        """Test Forward of ConvNet."""
        out = conv1D_net(torch.rand(size=[10, *conv1D_net.in_dim]))
        assert out.shape == (10, 64, 1)

    def test_forward_multi_input(self, multi_input_cnn):
        """Test Forward of Multi Input ConvNet."""
        master_device_setter(multi_input_cnn, 'cpu')
        input_tensor = [
            torch.rand(size=[10, 1, 28, 28]),
            torch.rand(size=[10, 1, 28, 28, 28]),
            [torch.rand(size=[10, 1, 28]),
             torch.rand(size=[10, 1, 28, 28])]
        ]
        out = multi_input_cnn(input_tensor)
        assert out.shape == (10, 10)

    def test_forward_pass_not_nan(self, conv3D_net):
        """Confirm out is non nan."""
        test_input = torch.rand(size=[1, *conv3D_net.in_dim])
        test_dataloader = DataLoader(TensorDataset(test_input, test_input))
        output = conv3D_net.forward_pass(
            data_loader=test_dataloader,
            transform_outputs=False)
        assert np.any(~np.isnan(output))

    def test_forward_pass_class_not_nan(self, conv3D_net_class):
        """Confirm out is non nan."""
        test_input = torch.rand(size=[1, *conv3D_net_class.in_dim])
        test_dataloader = DataLoader(TensorDataset(test_input, test_input))
        raw_output = conv3D_net_class.forward_pass(
            data_loader=test_dataloader,
            transform_outputs=False)
        class_output = conv3D_net_class.metrics.transform_outputs(
            in_matrix=raw_output)
        assert np.any(~np.isnan(class_output))
        assert np.any(~np.isnan(raw_output))

    def test_freeze_class(self, conv3D_net_class):
        """Test class network freezing."""
        conv3D_net_class.freeze(apply_inputs=False)
        for params in conv3D_net_class.network.parameters():
            assert params.requires_grad is False

    def test_unfreeze_class(self, conv3D_net_class):
        """Test class network unfreezing."""
        conv3D_net_class.freeze(apply_inputs=False)
        conv3D_net_class.unfreeze(apply_inputs=False)
        for params in conv3D_net_class.network.parameters():
            assert params.requires_grad is True

    def test_freeze_noclass(self, conv3D_net):
        """Test intermediate network freezing."""
        conv3D_net.freeze(apply_inputs=False)
        for params in conv3D_net.network.parameters():
            assert params.requires_grad is False

    def test_unfreeze_noclass(self, conv3D_net):
        """Test intermediate network unfreezing."""
        conv3D_net.freeze(apply_inputs=False)
        conv3D_net.unfreeze(apply_inputs=False)
        for params in conv3D_net.network.parameters():
            assert params.requires_grad is True

    def test_fit_multi_input(self, multi_input_cnn,
                             multi_input_cnn_train_loader,
                             multi_input_cnn_test_loader):
        """Test for fit function."""
        init_weights = copy.deepcopy(multi_input_cnn.network[0].
                                     _kernel.weight.detach())
        multi_input_cnn_no_fit = copy.deepcopy(multi_input_cnn)
        parameters1 = multi_input_cnn_no_fit.parameters()
        try:
            multi_input_cnn.fit(
                multi_input_cnn_train_loader,
                multi_input_cnn_test_loader,
                2)
        except RuntimeError:
            logger.error("The network multi_input_cnn failed to train.")
        finally:
            parameters2 = multi_input_cnn.parameters()
            trained_weights = multi_input_cnn.network[0]._kernel.weight.detach()

            # Sanity check if the network parameters are training
            assert not (torch.equal(init_weights.cpu(), trained_weights.cpu()))
            compare_params = [not torch.allclose(param1, param2)
                              for param1, param2 in zip(parameters1,
                                                        parameters2)]
            assert all(compare_params)

    def test_params_multi_input(self, multi_input_cnn,
                                multi_input_cnn_train_loader,
                                multi_input_cnn_test_loader):
        """Test for change in network params/specifications."""
        test_net = copy.deepcopy(multi_input_cnn)
        # Check the parameters are copying properly
        copy_params = [torch.allclose(param1, param2)
                       for param1, param2 in zip(multi_input_cnn.parameters(),
                                                 test_net.parameters())]
        assert all(copy_params)

        # Check the parameters change after copy and fit
        test_net.fit(
            multi_input_cnn_train_loader,
            multi_input_cnn_test_loader,
            2)
        close_params = [not torch.allclose(param1, param2)
                        for param1, param2 in zip(multi_input_cnn.parameters(),
                                                  test_net.parameters())]
        assert all(close_params)

        # Check the network params and optimizer params point to
        # the same memory
        if test_net.optim:
            assert isinstance(test_net.optim, torch.optim.Adam)
            assert isinstance(test_net.criterion, torch.nn.CrossEntropyLoss)
            for param, opt_param in zip(test_net.parameters(),
                                        test_net.optim.param_groups[0]['params']):
                assert param is opt_param

        # Check the params after saving loaading
        test_net.save_model()
        save_path = test_net.save_path
        abs_save_path = os.path.dirname(os.path.abspath(save_path))
        loaded_test_net = BaseNetwork.load_model(load_path=save_path)
        load_params = [torch.allclose(param1, param2)
                       for param1, param2 in zip(test_net.parameters(),
                                                 loaded_test_net.parameters())]
        shutil.rmtree(abs_save_path)
        assert all(load_params)

    def test_forward_pass_class_not_nan_single_value(self,
                                                     conv3D_net_class_single_value):
        """Confirm out is non nan."""
        test_input = torch.rand(size=[10, *conv3D_net_class_single_value.in_dim])
        test_output = torch.rand(size=[10, 1])
        test_dataloader = DataLoader(TensorDataset(test_input, test_output))
        raw_output = conv3D_net_class_single_value.forward_pass(
            data_loader=test_dataloader,
            transform_outputs=False)
        assert np.any(~np.isnan(raw_output))

    def test_early_stopping(self, conv3D_net_class_early_stopping,
                            conv3D_net_class):
        """ Test that their final params are different: aka
        that the early stopping did something.
        Constant seed resetting due to worker init function of dataloader
        https://github.com/pytorch/pytorch/issues/7068 """

        vulcanai.set_global_seed(42)

        ds = DataLoader(TensorDataset(
            torch.rand(size=[10, *conv3D_net_class_early_stopping.in_dim]),
            torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).long()),
            num_workers=0)

        vulcanai.set_global_seed(42)

        ds2 = copy.deepcopy(ds)

        vulcanai.set_global_seed(42)

        ds3 = copy.deepcopy(ds)

        conv3D_net_class_copy = copy.deepcopy(conv3D_net_class)

        conv3D_net_class_copy.fit(
            train_loader=ds2,
            val_loader=ds2,
            epochs=5)
        vulcanai.set_global_seed(42)

        conv3D_net_class.fit(
            train_loader=ds,
            val_loader=ds,
            epochs=5)

        vulcanai.set_global_seed(42)

        conv3D_net_class_early_stopping.fit(
            train_loader=ds3,
            val_loader=ds3,
            epochs=5)

        stopping_params = list(conv3D_net_class_early_stopping.parameters())[0][0][0].data
        non_stopping_params = list(conv3D_net_class.parameters())[0][0][0][0][0].data
        non_stopping_params_copy = list(conv3D_net_class_copy.parameters())[0][0][0][0][0].data

        assert not torch.eq(stopping_params, non_stopping_params).all()
        assert torch.eq(non_stopping_params, non_stopping_params_copy).all()




"""Specify dummy networks to test vulcan functionality."""
import pytest

import torch
from torch.utils.data import TensorDataset

from vulcanai.datasets import MultiDataset
from vulcanai.models import ConvNet, DenseNet


@pytest.fixture(scope="module")
def conv1D_net():
    """conv1D fixture."""
    return ConvNet(
        name='conv1D_net',
        in_dim=(1, 28),
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=24,
                    kernel_size=(5),
                    stride=2,
                    pool_size=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=24,
                    out_channels=64,
                    kernel_size=(5),
                    pool_size=2,
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def conv2D_net():
    """conv2D fixture."""
    return ConvNet(
        name='conv2D_net',
        in_dim=(1, 28, 28),
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=24,
                    kernel_size=(5, 5),
                    stride=2,
                    pool_size=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=24,
                    out_channels=64,
                    kernel_size=(5, 5),
                    pool_size=2,
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def conv3D_net():
    """conv3D fixture."""
    return ConvNet(
        name='conv3D_net',
        in_dim=(1, 28, 28, 28),
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=16,
                    kernel_size=(5, 5, 5),
                    stride=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=16,
                    out_channels=64,
                    kernel_size=(5, 5, 5),
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def conv3D_net_class():
    """conv3D fixture."""
    return ConvNet(
        name='conv3D_net',
        in_dim=(1, 28, 28, 28),
        num_classes=10,
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=16,
                    kernel_size=(5, 5, 5),
                    stride=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=16,
                    out_channels=64,
                    kernel_size=(5, 5, 5),
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )

@pytest.fixture(scope="module")
def conv3D_net_class_early_stopping():
    """conv3D fixture."""
    return ConvNet(
        name='conv3D_net',
        in_dim=(1, 28, 28, 28),
        num_classes=10,
        early_stopping="best_validation_error",
        early_stopping_patience=2,
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=16,
                    kernel_size=(5, 5, 5),
                    stride=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=16,
                    out_channels=64,
                    kernel_size=(5, 5, 5),
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )

@pytest.fixture(scope="module")
def conv3D_net_class_single_value():
    """conv3D fixture."""
    return ConvNet(
        name='conv3D_net',
        in_dim=(1, 28, 28, 28),
        num_classes=1,
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=16,
                    kernel_size=(5, 5, 5),
                    stride=2,
                    dropout=0.1
                ),
                dict(
                    in_channels=16,
                    out_channels=64,
                    kernel_size=(5, 5, 5),
                    dropout=0.1
                )
            ],
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def dnn_noclass():
    """DenseNet fixture."""
    return DenseNet(
        name='dnn_noclass',
        in_dim=(200),
        config={
            'dense_units': [100, 50],
            'dropout': [0.3, 0.5],
        }
    )


@pytest.fixture(scope="module")
def dnn_class():
    """DenseNet with prediction layer."""
    return DenseNet(
        name='dnn_class',
        in_dim=(200),
        config={
            'dense_units': [100, 50],
            'dropout': 0.5,
        },
        num_classes=3
    )


@pytest.fixture(scope="module")
def dnn_class_early_stopping():
    """DenseNet with prediction layer."""
    return DenseNet(
        name='dnn_class',
        in_dim=(200),
        early_stopping="best_validation_error",
        early_stopping_patience=2,
        config={
            'dense_units': [100, 50],
            'dropout': 0.5,
        },
        num_classes=3
    )


@pytest.fixture(scope="module")
def dnn_class_single_value():
    """DenseNet with prediction layer."""
    return DenseNet(
        name='dnn_class',
        in_dim=(200),
        config={
            'dense_units': [100, 50],
            'dropout': 0.5,
        },
        num_classes=1
    )

@pytest.fixture(scope="module")
def multi_input_dnn(conv1D_net, conv2D_net):
    """Dense network fixture with two inputs."""
    return DenseNet(
        name='multi_input_dnn',
        input_networks=[conv1D_net, conv2D_net],
        config={
            'dense_units': [100, 50],
            'initializer': None,
            'bias_init': None,
            'norm': None,
            'dropout': 0.5,  # Single value or List
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def multi_input_cnn(conv2D_net, conv3D_net, multi_input_dnn):
    """Bottom multi-input network fixture."""
    return ConvNet(
        name='multi_input_cnn',
        input_networks=[conv2D_net, conv3D_net, multi_input_dnn],
        num_classes=10,
        config={
            'conv_units': [
                dict(
                    in_channels=1,
                    out_channels=16,
                    kernel_size=(3, 3, 3),
                    stride=2,
                    dropout=0.1
                ),
            ],
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def multi_input_dnn_class(conv1D_net, conv2D_net):
    """Dense network fixture with two inputs."""
    return DenseNet(
        name='multi_input_dnn_class',
        input_networks=[conv1D_net, conv2D_net],
        num_classes=10,
        config={
            'dense_units': [100, 50],
            'initializer': None,
            'bias_init': None,
            'norm': None,
            'dropout': 0.5,  # Single value or List
        },
        device='cpu'
    )


@pytest.fixture(scope="module")
def multi_input_dnn_data(conv1D_net, conv2D_net,
                         multi_input_dnn):
    return MultiDataset([
        (
            TensorDataset(
                torch.rand(size=[10, *conv1D_net.in_dim]),
                torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).long()),
            True, True),
        (
            TensorDataset(
                torch.rand(
                    size=[
                        10,
                        *multi_input_dnn.input_networks['conv2D_net'].
                        in_dim])),
            True, False)
    ])


@pytest.fixture(scope="module")
def multi_input_cnn_data(conv2D_net, conv3D_net, multi_input_dnn_data):
    return MultiDataset([
        (TensorDataset(torch.rand(size=[10, *conv2D_net.in_dim])), True, False),
        (TensorDataset(torch.rand(size=[10, *conv3D_net.in_dim])), True, False),
        multi_input_dnn_data
    ])



"""Define the model utilitiy tests."""
import pytest
import numpy as np
import torch
from vulcanai.models.utils import (round_list,
                                    get_one_hot,
                                    pad,
                                    set_tensor_device,
                                    master_device_setter)

TEST_CUDA = torch.cuda.is_available()


def test_round_list():
    """Test if the list is rounded to the desired decimals."""
    test_list = [0.83754245, 0.13249807]
    out = round_list(test_list, decimals=3)
    assert len(str(out[0]).split(".")[1]) == 3


def test_get_one_hot():
    """Test for get_one_hot."""
    test_inp = np.array([0, 1, 2, 3, 4, 3, 2, 1, 0])
    assert np.all(get_one_hot(test_inp) == np.array([
                    [1., 0., 0., 0., 0.],
                    [0., 1., 0., 0., 0.],
                    [0., 0., 1., 0., 0.],
                    [0., 0., 0., 1., 0.],
                    [0., 0., 0., 0., 1.],
                    [0., 0., 0., 1., 0.],
                    [0., 0., 1., 0., 0.],
                    [0., 1., 0., 0., 0.],
                    [1., 0., 0., 0., 0.]
                ]))


def test_pad():
    """Test if input_tensor is padded to the desired padded shape."""
    n_channels = 2
    n_features = 50
    pad_shape = 250
    batch = 1
    test_inp = torch.randn([batch, n_channels, n_features])
    padded_tensor = pad(tensor=test_inp, target_shape=[pad_shape])
    assert padded_tensor.shape == (batch, n_channels, pad_shape)
    assert padded_tensor.nonzero().size(0) == n_channels * n_features
    # 2D padding
    pad_shape1 = 100
    test_inp = torch.randn([batch, n_channels, n_features, n_features])
    padded_tensor = pad(
        tensor=test_inp,
        target_shape=[pad_shape, pad_shape1])
    assert padded_tensor.shape == (batch, n_channels, pad_shape, pad_shape1)
    assert padded_tensor.nonzero().size(0) == \
        n_channels * n_features * n_features
    # 3D padding
    pad_shape2 = 50
    test_inp = torch.randn([batch, n_channels, n_features, n_features,
                            n_features])
    padded_tensor = pad(
        tensor=test_inp,
        target_shape=[pad_shape, pad_shape1, pad_shape2])
    assert padded_tensor.shape == \
        (batch, n_channels, pad_shape, pad_shape1, pad_shape2)
    assert padded_tensor.nonzero().size(0) == \
        n_channels * n_features * n_features * n_features


@pytest.mark.skipif(not TEST_CUDA, reason="No CUDA"
                    " supported devices available")
def test_set_tensor_device():
    """If CUDA is available, test set_tensor_device usage."""
    # Tensor
    test_tensor_cpu = torch.randn([5, 5]).cpu()
    test_tensor_cuda = set_tensor_device(test_tensor_cpu, device='cuda:0')
    assert str(test_tensor_cuda.device) == 'cuda:0'

    # List of Tensors
    test_tensorlist_cpu = [torch.randn([1, 2]).cpu(),
                           torch.randn([2, 3]).cpu(),
                           torch.randn([3, 4]).cpu()]
    test_tensorlist_cuda = set_tensor_device(
        test_tensorlist_cpu, device='cuda:0')
    for t in test_tensorlist_cuda:
        assert str(t.device) == 'cuda:0'


@pytest.mark.skipif(not TEST_CUDA, reason="No CUDA"
                    " supported devices available")
def test_master_device_setter(multi_input_cnn):
    """If CUDA is available, test master_device_setter usage."""
    # Make sure the network is in cpu first
    assert str(multi_input_cnn.device) == 'cpu'
    master_device_setter(multi_input_cnn, device='cuda:0')
    assert str(multi_input_cnn.device) == 'cuda:0'
    assert str(list(multi_input_cnn.input_networks.values())[0] == 'cuda:0')
    assert str(list(list(multi_input_cnn.input_networks.
                    values())[2].input_networks.values())[0] == 'cuda:0')



# -*- coding: utf-8 -*-
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

from .visualization import (
    compute_saliency_map,
    display_saliency_overlay,
    display_pca,
    display_tsne,
    display_confusion_matrix,
    display_record,
    display_receptive_fields
)

__all__ = [
    'utils',
    'visualization',
    'compute_saliency_map',
    'display_saliency_overlay',
    'display_pca',
    'display_tsne',
    'display_confusion_matrix',
    'display_record',
    'display_receptive_fields',
    ]

def start():
    " Start module"



# -*- coding: utf-8 -*-

from __future__ import print_function
import torch.utils.data as data
from PIL import Image
import os
import os.path
import errno
import torch
import codecs
import torchvision.transforms as transforms
import urllib
import io

# FROM https://raw.githubusercontent.com/mayurbhangale/fashion-mnist-pytorch/master/fashion.py
class FashionData(data.Dataset):
    """'MNIST <http://yann.lecun.com/exdb/mnist/>`_ Dataset.

    Parameters:
        root (string): Root directory of dataset where ``processed/training.pt``
            and  ``processed/test.pt`` exist.
        train (bool, optional): If True, creates dataset from ``training.pt``,
            otherwise from ``test.pt``.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
    """
    urls = [
        'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz',
        'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz',
        'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz',
        'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz',
    ]
    raw_folder = 'raw'
    processed_folder = 'processed'
    training_file = 'training.pt'
    test_file = 'test.pt'

    def __init__(self, root, train=True, transform=None, target_transform=None, download=False):
        self.root = os.path.expanduser(root)
        self.transform = transform
        self.target_transform = target_transform
        self.train = train  # training set or test set

        self.transform = transforms.Compose([transforms.ToTensor(),
                                            transforms.Normalize((0.1307,), (0.3081,))])

        if download:
            self.download()

        if not self._check_exists():
            print("downloading")
            self.download()

        if self.train:
            self.train_data, self.train_labels = torch.load(
                os.path.join(root, self.processed_folder, self.training_file))
        else:
            self.test_data, self.test_labels = torch.load(os.path.join(root, self.processed_folder, self.test_file))

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        if self.train:
            img, target = self.train_data[index], self.train_labels[index]
        else:
            img, target = self.test_data[index], self.test_labels[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.numpy(), mode='L')

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        if self.train:
            return len(self.train_data)
        else:
            return len(self.test_data)

    def _check_exists(self):
        return os.path.exists(os.path.join(self.root, self.processed_folder, self.training_file)) and \
            os.path.exists(os.path.join(self.root, self.processed_folder, self.test_file))

    # TODO: Need to fix. File not found error before it downloads
    def download(self):
        """Download the MNIST data if it doesn't exist in processed_folder already."""
        import gzip

        if self._check_exists():
            return

        # download files
        try:
            os.makedirs(os.path.join(self.root, self.raw_folder))
            os.makedirs(os.path.join(self.root, self.processed_folder))
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise

        for url in self.urls:
            print('Downloading ' + url)
            response = urllib.request.urlopen(url)
            #compressed_file = io.BytesIO(response.read())
            filename = url.rpartition('/')[2]
            file_path = os.path.join(self.root, self.raw_folder, filename)
            with open(file_path.replace('.gz', ''), 'wb') as out_f, \
                    gzip.GzipFile(fileobj=response) as zip_f:
                out_f.write(zip_f.read())
            #os.unlink(file_path)

        # process and save as torch files
        print('Processing...')

        training_set = (
            read_image_file(os.path.join(self.root, self.raw_folder, 'train-images-idx3-ubyte')),
            read_label_file(os.path.join(self.root, self.raw_folder, 'train-labels-idx1-ubyte'))
        )
        test_set = (
            read_image_file(os.path.join(self.root, self.raw_folder, 't10k-images-idx3-ubyte')),
            read_label_file(os.path.join(self.root, self.raw_folder, 't10k-labels-idx1-ubyte'))
        )
        with open(os.path.join(self.root, self.processed_folder, self.training_file), 'wb') as f:
            torch.save(training_set, f)
        with open(os.path.join(self.root, self.processed_folder, self.test_file), 'wb') as f:
            torch.save(test_set, f)

        print('Done!')


def get_int(b):
    return int(codecs.encode(b, 'hex'), 16)


def parse_byte(b):
    if isinstance(b, str):
        return ord(b)
    return b


def read_label_file(path):
    with open(path, 'rb') as f:
        data_in = f.read()
        assert get_int(data_in[:4]) == 2049
        length = get_int(data_in[4:8])
        labels = [parse_byte(b) for b in data_in[8:]]
        assert len(labels) == length
        return torch.LongTensor(labels)


def read_image_file(path):
    with open(path, 'rb') as f:
        data_in = f.read()
        assert get_int(data_in[:4]) == 2051
        length = get_int(data_in[4:8])
        num_rows = get_int(data_in[8:12])
        num_cols = get_int(data_in[12:16])
        images = []
        idx = 16
        for l in range(length):
            img = []
            images.append(img)
            for r in range(num_rows):
                row = []
                img.append(row)
                for c in range(num_cols):
                    row.append(parse_byte(data_in[idx]))
                    idx += 1
        assert len(images) == length
        return torch.ByteTensor(images).view(-1, 28, 28)



"""Setup script for uploading package to PyPI servers."""
from setuptools import setup, find_packages

tests_require = [
    'pytest',
    'numpydoc'
]

docs_require = [
    'Sphinx', # TODO: maybe numpydoc?
]

with open('requirements.txt') as f:
    install_requires = [l.strip() for l in f]


setup(
    name='vulcanai',
    version='1.0.7',
    description='A high-level framework built on top of Pytorch'
                ' using added functionality from Scikit-learn to provide '
                'all of the tools needed for visualizing and processing '
                'high-dimensional data, modular neural networks, '
                'and model evaluation',
    author='Robert Fratila, Priyatharsan Rajasekar, Caitrin Armstrong, '
            'Joseph Mehltretter',
    author_email='robertfratila10@gmail.com',
    url='https://github.com/Aifred-Health/Vulcan',
    install_requires=install_requires,
    extras_require={
        'testing': tests_require,
        'docs': docs_require,
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=['Development Status :: 3 - Alpha',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Education',
                 'Topic :: Software Development :: Build Tools',
                 'Programming Language :: Python :: 3.6',
                 'Operating System :: Unix',
                 'Operating System :: POSIX :: Linux',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence'],
    keywords='deep learning machine learning development framework'
)



"""Contains all ensemble models."""
from copy import deepcopy
import logging
from datetime import datetime
import pickle
import os

import torch
from torch import nn
from torch.optim.lr_scheduler import CosineAnnealingLR

from .basenetwork import BaseNetwork

logger = logging.getLogger(__name__)


class SnapshotNet(BaseNetwork):
    """
    Initialize snapshot ensemble given a template network.

    A wrapper class for any Network inheriting from BaseNetwork to
    train the template network using Snapshot Ensembling.

    Parameters:
        name : str
            String of snapshot ensemble name.
        template_network : BaseNetwork
            Network object which you want to ensemble.
        n_snapshots : int
            Number of snapshots in ensemble.

    Returns:
        network : SnapshotNet

    """

    def __init__(self, name, template_network, n_snapshots=3):
        """Use Network to build model snapshots."""
        # TODO: Should these be defaulted to the values of template_network?
        super(SnapshotNet, self).__init__(
            name=name,
            config=None,  # template_network._config
            in_dim=template_network.in_dim,
            save_path=None,  # template_network.save_path
            input_networks=None,  # template_network.input_networks
            num_classes=template_network._num_classes,
            activation=None,  # template_network.network[0]._activation
            pred_activation=None,  # pred_activation
            optim_spec=None,  # template_network._optim_spec
            lr_scheduler=None,  # template_network.lr_scheduler
            early_stopping=None,  # template_network.early_stopping
            criter_spec=None  # template_network._criter_spec
        )

        if not isinstance(template_network, BaseNetwork):
            raise ValueError(
                "template_network type must inherit from BaseNetwork.")

        self.template_network = deepcopy(template_network)
        self.network = nn.ModuleList()
        self.out_dim = self.template_network.out_dim
        if n_snapshots <= 0:
            raise ValueError("n_snapshots must be >=1.")
        self.n_snapshots = n_snapshots

    def fit(self, train_loader, val_loader, epochs,
            retain_graph=None, valid_interv=4, plot=False):
        """
        Train each model for T/M epochs and controls network learning rate.

        Collects each model in a class variable self.network

        Parameters:
            train_loader : DataLoader
                Input data and targets to train against
            val_loader : DataLoader
                Input data and targets to validate against
            epochs : int
                Total number of epochs (evenly distributed between snapshots)

        Returns:
            None

        """
        # There must be at least one train epoch for each snapshot
        if epochs < self.n_snapshots:
            logger.warn(
                'Number of epochs to small for number of Snapshots. '
                'Setting epochs to {}.'.format(self.n_snapshots))
            epochs = self.n_snapshots

        T = epochs
        # How many epochs each singular network should train for
        network_epochs = T // self.n_snapshots

        # Temporary but check if it first has an optimizer,
        # if not it will make one
        if self.template_network.optim is None:
            self.template_network._init_trainer()

        self.template_network.lr_scheduler = CosineAnnealingLR(
            optimizer=self.template_network.optim,
            T_max=network_epochs
        )

        for index in range(self.n_snapshots):
            self.template_network.fit(
                train_loader=train_loader,
                val_loader=val_loader,
                epochs=network_epochs,
                valid_interv=valid_interv,
                plot=plot
            )
            # Save instance of snapshot in a nn.ModuleList
            temp_network = deepcopy(self.template_network)
            self._update_network_name_stack(
                network=temp_network,
                append_str=index)
            self.network.append(temp_network)

    def _update_network_name_stack(self, network, append_str):
        """
        Given a network, append a string to the name of all networks in stack.

        Recursively traverse each input network to update names
        with the appended string.

        Parameters:
            network : BaseNetwork
                Network stack to update names of with new append_str.
            append_str : int, str
                The characters to append at the end of BaseNetwork stack of names.

        """
        if network.input_networks:
            for in_net in network.input_networks.values():
                self._update_network_name_stack(in_net, append_str)
        network.name = "{}_{}".format(network.name, append_str)

    def forward(self, inputs, **kwargs):
        """
        Snapshot forward function.

        Collect outputs of all internal networks and average outputs.

        Parameters:
            inputs : torch.Tensor
                Input tensor to pass through self.

        Returns:
            output : torch.Tensor

        """
        if len(self.network) == 0:
            raise ValueError("SnapshotNet needs to be trained.")

        pred_collector = []
        for net in self.network:
            pred_collector.append(net(inputs))
        # Stack outputs along a new 0 dimension to be averaged
        pred_collector = torch.stack(pred_collector)

        return torch.mean(input=pred_collector, dim=0)

    def save_model(self, save_path=None):
        """
        Save all ensembled network in a folder with ensemble name.

        Parameters:
            save_path : str
                The folder path to save models in.

        Returns:
            None

        """
        if not save_path:
            save_path = r"saved_models/"

        if not save_path.endswith("/"):
            save_path = save_path + "/"

        save_path = save_path + "{}_{}/".format(
            self.name, datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        logger.info("No save path provided, saving to {}".format(save_path))

        for network in self.network:
            logger.info("Saving network {}".format(network.name))
            network.save_model(save_path=save_path)

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        model_save_path = save_path + "model.pkl"
        self.save_path = save_path
        pickle.dump(self, open(model_save_path, "wb"), 2)



"""Define utilities for all networks."""
from math import ceil, floor
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch.utils.data import DataLoader, TensorDataset

import numpy as np
import math
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelBinarizer
from collections import OrderedDict as odict
from collections import defaultdict

def _get_probs(network, loader, index_to_iter, ls_feat_vals):
    """
    Returns probability for each object within loader based on output from training neural network

    Parameters:
        network : vulcan.model
	    training vulcan network
	loader : torch.dataloader
	    dataloader containing validation set
	index_to_iter : string
	    feature to iterate through adjusting values
	ls_feat_vals : list
	    values to iterate through for feature in index_to_iter
    Returns:
	dct_scores : dictionary
	    dictionary of scores
    """
    dct_scores = defaultdict()
    for index in range(len(loader)):
        dct_scores[index] = {}
    for index in range(len(loader)):
        # Extract specific index from loader and create a DataLoader instance to send to forward_pass
        input_loader = DataLoader(TensorDataset(loader.dataset[index][0].unsqueeze(0), \
                                                loader.dataset[index][1].unsqueeze(0)))
        
        subj_prob = network.forward_pass(data_loader=input_loader, transform_outputs=False)
        # Standardize probability of positive label
        subj_prob = subj_prob[0][1] * 100
        subj_prob = round(subj_prob, 2)
        # Add probability to scores dictionary where keys are the index and value the probability belongs to.
        dct_scores[index][loader.dataset[index][0][index_to_iter].item()] = subj_prob
        # Iterate through other possible values and find probability of positive label and add to dictionary
        # for current index.
        for new_val in ls_feat_vals:
            if new_val != loader.dataset[index][0][index_to_iter].item():
                loader.dataset[index][0][index_to_iter] = new_val
                input_loader = DataLoader(TensorDataset(loader.dataset[index][0].unsqueeze(0), \
                                                        loader.dataset[index][1].unsqueeze(0)))
                subj_prob = network.forward_pass(data_loader=input_loader, transform_outputs=False)
                subj_prob = subj_prob[0][1] * 100
                subj_prob = round(subj_prob, 2)
                dct_scores[index][new_val] = subj_prob
    return dct_scores

def _filter_matched_subj(dct_scores, loader, index_to_iter):
    """
    Returns dictionary of filtered keys based on predicted value and value truly assigned in actual set.

    Parameters:
        dct_scores : dictionary
            dictionary of probability scores produced in get_scores function
        loader : torch.dataloader
            dataloader containing validation set
        index_to_iter : string
            feature to iterate through adjusting values
    Returns:
        dct_filtered : dictionary
            dictionary of filtered entities 
    """
    dct_filtered = {}
    for subj in list(dct_scores):
        highest_prob = max(dct_scores[subj].values())
        highest_val = [val for val, prob in dct_scores[subj].items() if prob==highest_prob]
        if loader.dataset.dataset[subj][0][index_to_iter].item() in highest_val:
            dct_filtered[subj] = highest_prob
    return dct_filtered

def round_list(raw_list, decimals=4):
    """
    Return the same list with each item rounded off.

    Parameters:
        raw_list : float list
            float list to round.
        decimals : int
            How many decimal points to round to.

    Returns:
        rounded_list : float list
            The rounded list in the same shape as raw_list.

    """
    return [round(item, decimals) for item in raw_list]


def get_one_hot(in_matrix):
    """
    Reformat truth matrix to same size as the output of the dense network.

    Parameters:
        in_matrix : numpy.ndarray
            The categorized 1D matrix

    Returns:
        one_hot : numpy.ndarray
            A one-hot matrix representing the categorized matrix

    """
    if in_matrix.dtype.name == 'category':
        custom_array = in_matrix.cat.codes

    elif isinstance(in_matrix, np.ndarray):
        custom_array = in_matrix

    else:
        raise ValueError("Input matrix cannot be converted.")

    lb = LabelBinarizer()
    return np.array(lb.fit_transform(custom_array), dtype='float32')


def pad(tensor, target_shape):
    """
    Pad incoming tensor to the size of target_shape.

    tensor must have same spatial dimenison as the target_shape.
    Useful for combining various conv dimension outputs and to implement
    'same' padding for conv operations.

    Parameters:
        tensor : torch.Tensor
            Tensor to be padded
        target_shape : np.array
            Final padded tensor shape [*spatial_dimensions]

    Returns:
        tensor : torch.Tensor
            zero padded tensor with spatial dimension as target_shape

    """
    # Ignore channels and batch and focus on spatial dimensions
    # from incoming tensor
    if not isinstance(target_shape, np.ndarray):
        target_shape = np.array(target_shape)
    n_dim = len(target_shape)
    # Calculate, element-wise, how much needs to be padded for each dim.
    dims_size_diff = target_shape - np.array(tensor.shape[-n_dim:])
    # TODO: Use torch.nn.ConstantPadding?
    padding_needed = []
    for dim_diff in reversed(dims_size_diff):
        dim_zero_padding = ceil(dim_diff/2)
        dim_one_padding = floor(dim_diff/2)
        padding_needed.append(dim_zero_padding)
        padding_needed.append(dim_one_padding)
    return F.pad(tensor, padding_needed)


def network_summary(network, input_size=None):
    """
    Returns the summary of shapes of all layers in the network
    :return: OrderedDict of shape of each layer in the network
    """
    if not input_size:
        input_size=[]
        for net in network.input_networks.values():
            input_size.append(net.in_dim)
    # input_size must be a list
    if isinstance(input_size, (tuple, int)):
        input_size = [input_size]
    
    def get_size(summary_dict, output):
        """
        Helper function for the BaseNetwork's get_output_shapes
        """
        if isinstance(output, tuple):
            for i in range(len(output)):
                summary_dict[i] = odict()
                summary_dict[i] = get_size(summary_dict[i], output[i])
        else:
            summary_dict['output_shape'] = tuple(output.size())
        return summary_dict

    def register_hook(module):
        """
        Registers a backward hook
        For more info: https://pytorch.org/docs/stable/_modules/torch/tensor.html#Tensor.register_hook
        """
        def hook(module, input, output):
            """
            https://github.com/pytorch/tutorials/blob/8afce8a213cb3712aa7de1e1cf158da765f029a7/beginner_source/former_torchies/nn_tutorial.py#L146
            """
            class_name = str(module.__class__).split('.')[-1].split("'")[0]
            module_idx = len(summary)
            # Test
            m_key = '%s-%i' % (class_name, module_idx + 1)
            summary[m_key] = odict()
            summary[m_key]['input_shape'] = tuple(input[0].size())
            summary[m_key] = get_size(summary[m_key], output)
            # Test
            params = 0
            if hasattr(module, 'weight'):
                params += torch.prod(torch.LongTensor(tuple(module.weight.size())))
                if module.weight.requires_grad:
                    summary[m_key]['trainable'] = True
                else:
                    summary[m_key]['trainable'] = False
            if hasattr(module, 'bias'):
                params += torch.prod(torch.LongTensor(tuple(module.bias.size())))
            # Test
            summary[m_key]['nb_params'] = params
        if not isinstance(module, torch.nn.Sequential) and \
                not isinstance(module, torch.nn.ModuleList) and \
                not (module == network):
            hooks.append(module.register_forward_hook(hook))

    x = []
    for in_size in input_size:
        x.append(torch.empty(1, *in_size))
    
    if len(x)==1:
        x = torch.cat(x,dim=1)

    # create properties
    summary = odict()
    hooks = []

    # register hook
    network.apply(register_hook)
    # make a forward pass
    network.cpu()(x)

    # remove these hooks
    for h in hooks:
        h.remove()

    return summary


def print_model_structure(network, input_size=None):
    """Print the entire model structure."""
    shapes = network_summary(network)
    for k, v in shapes.items():
        print('{}:'.format(k))
        if isinstance(v, odict):
            for k2, v2 in v.items():
                print('\t {}: {}'.format(k2, v2))


def selu_weight_init_(tensor, mean=0.0):
    """
    SELU layer weight initialization function.

    Function assigned to variable that will be called within
    _init_weights function to assign weights for selu.

    Parameters:
        tensor :  torch.tensor
            Weight tensor to be adjusted
        mean : float
            Mean value for the normal distribution

    Returns:
        torch.tensor
            weight tensor with normal distribution

    """
    with torch.no_grad():
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(tensor)
        std = math.sqrt(1. / fan_in)
        return nn.init.normal_(tensor, mean, std)


def selu_bias_init_(tensor, const=0.0):
    """
    SELU layer bias initialization function.

    Function assigned to variable that will be called within
    _init_bias function to assign bias for selu.

    Parameters:
        tensor : torch.tensor
            Bias tensor to be adjusted
        const : float
            Constant value to be assigned to tensor.

    Returns:
        torch.tensor
            bias tensor with constant values.

    """
    with torch.no_grad():
        return nn.init.constant_(tensor, const)


def set_tensor_device(data, device=None):
    """
    Convert list of data tensors to specified device.

    Parameters:
        data : torch.tensor or list
            data to be converted to the specified device.
        device : str or torch.device
            the desired device

    Returns:
        data : torch.tensor or list
            data converted to the specified device

    """
    if not isinstance(data, (list, tuple)):
        data = data.to(device=device, non_blocking=True)

    else:
        for idx, d in enumerate(data):
            data[idx] = set_tensor_device(d, device=device)
    return data


def master_device_setter(network, device=None):
    """
    Convert network and input_networks to specified device.

    Parameters:
        network : BaseNetwork
            network to be converted to the specified device.
        device : str or torch.device
            the desired device

    """
    network.device = device
    if network.input_networks:
        for net in network.input_networks.values():
            master_device_setter(net, device)



"""Will test visualization functions."""
import pytest
import numpy as np
import pandas as pd
import torch
from torch.utils.data import TensorDataset, DataLoader
from vulcanai.datasets import tabulardataset
import os
from copy import deepcopy

from sklearn.datasets import load_digits
from sklearn.metrics import confusion_matrix
from vulcanai.plotters.visualization import compute_saliency_map, \
                                            display_pca, \
                                            display_tsne, \
                                            display_receptive_fields, \
                                            display_saliency_overlay, \
                                            display_confusion_matrix


class TestVisualization:
    """Test all visualization functionality."""

    @pytest.fixture
    def cnn_class(self):
        """Create ConvNet with classes fixture."""
        from vulcanai.models.cnn import ConvNet
        return ConvNet(
            name='Test_ConvNet_class',
            in_dim=(1, 28, 28),
            config={
                'conv_units': [
                    {
                        "in_channels": 1,
                        "out_channels": 16,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    },
                    {
                        "in_channels": 16,
                        "out_channels": 1,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    }]
            },
            num_classes=3
        )

    @pytest.fixture
    def dnn_class(self):
        """Create DenseNet with classes fixture."""
        from vulcanai.models.dnn import DenseNet
        return DenseNet(
            name='Test_DenseNet_class',
            in_dim=(200),
            config={
                'dense_units': [100, 50],
                'dropouts': 0.3,
            },
            num_classes=3
        )

    @pytest.fixture
    def dnn_class_two(self):
        """Create DenseNet with no prediction layer."""
        from vulcanai.models.dnn import DenseNet
        return DenseNet(
            name='Test_DenseNet_class',
            in_dim=(200),
            activation=torch.nn.SELU(),
            num_classes=2,
            input_networks=None,
            config={
                'dense_units': [100],
                'dropout': [0.3],
                'initializer': None,
                'bias_init': None,
                'norm': None
            },
            optim_spec={'name': 'Adam', 'lr': 0.001}
        )

    def test_compute_saliency_map_cnn(self, cnn_class):
        """Confirm hooks are removed, and gradient shape."""
        test_input_1B = torch.ones([1, *cnn_class.in_dim])
        test_input_5B = torch.ones([5, *cnn_class.in_dim])

        model_copy = deepcopy(cnn_class)
        # Test shape conservation
        cnn_class.freeze(apply_inputs=False)
        sal_map_1B = compute_saliency_map(
            cnn_class,
            test_input_1B, torch.LongTensor([2]))
        for sal_map, test_input in zip(sal_map_1B, [test_input_1B]):
            assert sal_map.shape == test_input.shape

        cnn_class.unfreeze(apply_inputs=False)
        sal_map_5B = compute_saliency_map(
            cnn_class, test_input_5B,
            torch.LongTensor([0, 2, 1, 1, 0]))
        for sal_map, test_input in zip(sal_map_5B, [test_input_5B]):
            assert sal_map.shape == test_input.shape

        # Check that all gradients are not 0
        assert ~np.all(sal_map_5B == 0.)

        # Test hook removal
        assert cnn_class._backward_hooks == model_copy._backward_hooks

    def test_compute_saliency_map_dnn(self, dnn_class):
        """Confirm hooks are removed, and gradient shape."""
        test_input_1B = torch.ones([1, *dnn_class.in_dim])
        test_input_5B = torch.ones([5, *dnn_class.in_dim])
        model_copy = deepcopy(dnn_class)
        # Test shape conservation
        dnn_class.freeze(apply_inputs=False)
        sal_map_1B = compute_saliency_map(
            dnn_class,
            test_input_1B, torch.LongTensor([2]))
        for sal_map, test_input in zip(sal_map_1B, [test_input_1B]):
            assert sal_map.shape == test_input.shape

        dnn_class.unfreeze(apply_inputs=False)
        sal_map_5B = compute_saliency_map(
            dnn_class, test_input_5B,
            torch.LongTensor([0, 2, 1, 1, 0]))
        for sal_map, test_input in zip(sal_map_5B, [test_input_5B]):
            assert sal_map.shape == test_input.shape

        # Check that all gradients are not 0
        assert ~np.all(sal_map_5B == 0.)

        # Test hook removal
        assert dnn_class._backward_hooks == model_copy._backward_hooks

    def test_display_saliency_overlay(self, dnn_class):
        """Test saliency overlay displays and saves."""
        curr_path = str(os.path.dirname(__file__)) + '/'
        img = np.zeros((25,25))
        test_input_1B = torch.ones([1, *dnn_class.in_dim])
        sal_map_1B = np.array(compute_saliency_map(
            dnn_class,
            test_input_1B, torch.LongTensor([2])))
        display_saliency_overlay(img, sal_map_1B, shape=(25,25), save_path=curr_path)
        file_created = False
        for file in os.listdir(curr_path):
            if file.startswith('saliency'):
                file_created = True
                file_path = curr_path + file
                os.remove(file_path)
        assert file_created

    def test_display_pca(self, dnn_class):
        """Test PCA displays and saves."""
        curr_path = str(os.path.dirname(__file__)) + '/'
        digits = load_digits()
        display_pca(
            digits.data[0:10], digits.target[0:10], save_path=curr_path)
        file_created = False
        for file in os.listdir(curr_path):
            if file.startswith('PCA'):
                file_created = True
                file_path = curr_path + file
                os.remove(file_path)
        assert file_created

    def test_display_tsne(self, dnn_class):
        """Test t-SNE displays and saves."""
        curr_path = str(os.path.dirname(__file__)) + '/'
        digits = load_digits()
        display_tsne(
            digits.data[0:10], digits.target[0:10], save_path=curr_path)
        file_created = False
        for file in os.listdir(curr_path):
            if file.startswith('t-SNE'):
                file_created = True
                file_path = curr_path + file
                os.remove(file_path)
        assert file_created

    def test_display_confusion_matrix(self):
        """Test confusion matrix displays and saves."""
        curr_path = str(os.path.dirname(__file__)) + '/'
        cm = confusion_matrix(y_true=[0, 1, 1, 0, 0], y_pred=[1, 1, 0, 1, 0])
        display_confusion_matrix(cm, class_list=[0, 1], save_path=curr_path)
        file_created = False
        for file in os.listdir(curr_path):
            if file.startswith('confusion'):
                file_created = True
                file_path = curr_path + file
                os.remove(file_path)
        assert file_created

    def test_receptive_field(self, dnn_class_two):
        """Test receptive field visualization gets created and can be saved."""
        curr_path = str(os.path.dirname(__file__)) + '/'
        test_input = torch.ones([10, *dnn_class_two.in_dim]).float()
        test_target = torch.LongTensor([0, 1, 1, 0, 0, 1, 0, 1, 0, 1])
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))

        dnn_class_two.fit(test_dataloader, test_dataloader, 5)
        display_receptive_fields(dnn_class_two, save_path=curr_path)
        file_created = False
        for file in os.listdir(curr_path):
            if file.startswith('feature_importance'):
                file_created = True
                file_path = curr_path + file
                os.remove(file_path)
        assert file_created



""" Imports network classes so they can be used directly"""
from .utils import (
        round_list,
        get_one_hot,
        pad,
        network_summary,
        print_model_structure,
        selu_weight_init_,
        selu_bias_init_,
        set_tensor_device,
        master_device_setter
    )

from .basenetwork import BaseNetwork
from .cnn import ConvNet, ConvNetConfig
from .dnn import DenseNet, DenseNetConfig
from .ensemble import SnapshotNet
from .metrics import Metrics
from .layers import BaseUnit, ConvUnit, DenseUnit, FlattenUnit

__all__ = [
    'basenetwork',
    'cnn',
    'dnn',
    'layers',
    'ensemble',
    'metrics',
    'utils',
    'BaseNetwork',
    'ConvNet',
    'ConvNetConfig',
    'DenseNet',
    'DenseNetConfig',
    'SnapshotNet',
    'Metrics',
    'ConvUnit',
    'DenseUnit',
    'FlattenUnit',
    'BaseUnit'
]

def start():
    " Start module"



# coding=utf-8
""" Defines the MultiDataset Class"""
from torch.utils.data import Dataset
import logging

logger = logging.getLogger(__name__)


class MultiDataset(Dataset):
    """
    Define a dataset for multi input networks.

    Takes in a list of datasets, and whether or not their input_data
    and target data should be output.

    Parameters:
        dataset_tuples : list of tuples
            Each tuple being (Dataset, use_data_boolean, use_target_boolean).
            A list of tuples, wherein each tuple should have the Dataset in the
            zero index, a boolean of whether to include the input_data in
            the first index, and a boolean of whether to include the target
            data in the second index. You can only specificy one target at a
            time throughout all incoming datasets.

    Returns:
        multi_dataset : torch.utils.data.Dataset

    """

    def __init__(self, dataset_tuples):
        """Initialize a dataset for multi input networks."""
        def get_total_targets(multi_datasets):
            num_targets = 0
            for tup in multi_datasets:
                if isinstance(tup, MultiDataset):
                    num_targets += get_total_targets(tup._dataset_tuples)
                else:
                    num_targets += int(tup[2])
            return num_targets

        self._dataset_tuples = dataset_tuples
        # must always have exactly one target.
        total_num_targets = get_total_targets(self._dataset_tuples)
        if total_num_targets > 1:
            raise ValueError(
                "You may specify at most one target."
                " {} specified".format(total_num_targets))

    def __len__(self):
        """
        Denotes the total number of samples.

        Will look for the dataset with the smallest number of samples and
        default the length to that so as to avoid getting a sample that doesn't
        exist in another dataset.

        Returns:
            length : int

        """
        logger.warning("Defaulting to the length of the smallest dataset")

        def get_min_length(multi_datasets):
            min_length = float('inf')
            for tup in multi_datasets:
                if isinstance(tup, MultiDataset):
                    length = get_min_length(tup._dataset_tuples)
                else:
                    length = len(tup[0])
                if length < min_length:
                    min_length = length
            return min_length

        return get_min_length(self._dataset_tuples)

    def __getitem__(self, idx):
        """
        Override getitem used by DataLoader required by torch Dataset.

        Parameters:
            idx : index
                Index of sample to extract

        Returns:
            (input_data, targets) : (torch.Tensor, torch.Tensor)
                Tuple of input_data and target at that index as specific in
                config. Target data is always last.

        """
        input_data_items = []
        target_item = None

        for tup in self._dataset_tuples:

            include_data = tup[1]
            include_target = tup[2]

            if isinstance(tup, MultiDataset):
                ds = tup
            else:
                ds = tup[0]

            if include_data:
                input_data_items.append(ds.__getitem__(idx)[0])

            if include_target:
                target_item = ds.__getitem__(idx)[1]

        return input_data_items, target_item



"""Define the ConvUnit and DenseUnit."""
import torch
import torch.nn as nn
import logging
from .utils import selu_bias_init_, selu_weight_init_
logger = logging.getLogger(__name__)


class BaseUnit(nn.Sequential):
    """
    The base class for all layers.

    Parameters:
        weight_init : torch.nn.init
            Torch initialization function for weights.
        bias_init : int or float
            A constant int or float to initialize biases with.
        norm : str
            'batch' for batch norm of 'instance' for instance norm.
        dropout : float between 0-1
            The probability of dropping out a feature during training.

    Returns:
        dense_unit : torch.nn.Sequential
            A single fully connected layer.

    """

    def __init__(self, weight_init=None, bias_init=None,
                 norm=None, dropout=None):
        """Initialize a base unit."""
        super(BaseUnit, self).__init__()

        self.weight_init = weight_init
        self.bias_init = bias_init
        self.norm = norm
        self.dropout = dropout

        self.in_shape = None  # [self.batch_size, *in_shape]
        self.out_shape = None  # [self.batch_size, *out_shape]
        self.in_bound_layers = []
        self.out_bound_layers = []

        self._kernel = None

    def _init_weights(self):
        """
        Initialize the weights.

        If self.weight_init is None, then pytorch default weight
        will be assigned to the kernel.
        """
        if self.weight_init:
            self.weight_init(self._kernel.weight)

    def _init_bias(self):
        """
        Initialize the bias.

        If self.bias_init is None, then pytorch default bias
        will be assigned to the kernel.
        """
        if self.bias_init:
            self.bias_init(self._kernel.bias)


class FlattenUnit(BaseUnit):
    """
    Layer to flatten the input.

    Returns:
        flatten_unit : torch.Sequential
            A flatten layer.

    """

    def __init__(self):
        """Initialize flatten layer."""
        super(FlattenUnit, self).__init__()

    def forward(self, x):
        """Maintain batch size but flatten all remaining dimensions."""
        return x.view(x.shape[0], -1)


class DenseUnit(BaseUnit):
    """
    Define the DenseUnit object.

    Parameters:
        in_features : int
            The incoming feature size of a sample.
        out_features : int
            The number of hidden Linear units for this layer.
        weight_init : torch.nn.init
            Torch initialization function.
        bias_init : int or float
            A constant int or float to initialize biases with.
        norm : str
            'batch' for batch norm of 'instance' for instance norm.
        activation : torch.nn.Module
            An activation function to apply after Linear unit.
        dropout : float between 0-1
            The probability of dropping out a feature during training.

    Returns:
        dense_unit : torch.nn.Sequential
            A single fully connected layer.

    """

    def __init__(self, in_features, out_features,
                 weight_init=None, bias_init=None,
                 norm=None, activation=None, dropout=None):
        """Initialize a single DenseUnit (i.e. a dense layer)."""
        super(DenseUnit, self).__init__(weight_init, bias_init,
                                        norm, dropout)
        self.in_features = in_features
        self.out_features = out_features
        # Main layer
        self._kernel = nn.Linear(
                            in_features=self.in_features,
                            out_features=self.out_features,
                            bias=True
                            )
        self.add_module('_kernel', self._kernel)

        # Norm
        if self.norm is not None:
            if self.norm == 'batch':
                self.add_module(
                    '_norm',
                    torch.nn.BatchNorm1d(self.out_features))
            elif self.norm == 'instance':
                self.add_module(
                    '_norm',
                    torch.nn.InstanceNorm1d(self.out_features))

        # Activation/Non-Linearity
        if activation is not None:
            self.add_module('_activation', activation)
            if isinstance(activation, nn.SELU):
                self.weight_init = selu_weight_init_
                self.bias_init = selu_bias_init_

        # Dropout
        if self.dropout is not None:
            if isinstance(activation, nn.SELU):
                self.add_module(
                    '_dropout', nn.AlphaDropout(self.dropout))
            else:
                self.add_module(
                    '_dropout', nn.Dropout(self.dropout))
        self._init_weights()
        self._init_bias()


# TODO: Automatically calculate padding to be the same as input shape.
class ConvUnit(BaseUnit):
    """
    Define the ConvUnit object.

    Parameters:
        conv_dim : int
            1, 2, or 3 representing spatial dimensional inputs.
        in_channels : int
            The number of incoming channels.
        out_channels : int
            The number of convolution kernels for this layer.
        kernel_size : int or tuple
            The size of the 1, 2, or 3 dimensional convolution kernel.
        weight_init : torch.nn.init
            Torch initialization function.
        bias_init : int or float
            A constant int or float to initialize biases with.
        stride : int or tuple
            The stride of the 1, 2, or 3 dimensional convolution kernel.
        padding : int
            Number of zero-padding on both sides per dimension.
        norm : str
            'batch' for batch norm of 'instance' for instance norm.
        activation : torch.nn.Module
            An activation function to apply after Linear unit.
        pool_size : int
            Max pooling by a factor of pool_size in each dimension.
        dropout : float between 0-1
            The probability of dropping out a feature during training.

    Returns:
        conv_unit : torch.nn.Sequential
            A single convolution layer.

    """

    def __init__(self, conv_dim, in_channels, out_channels, kernel_size,
                 weight_init=None, bias_init=None,
                 stride=1, padding=0, norm=None,
                 activation=None, pool_size=None, dropout=None):
        """Initialize a single ConvUnit (i.e. a conv layer)."""
        super(ConvUnit, self).__init__(weight_init, bias_init,
                                       norm, dropout)
        self.conv_dim = conv_dim
        self._init_layers()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size

        # Main layer
        self._kernel = self.conv_layer(
                              in_channels=self.in_channels,
                              kernel_size=self.kernel_size,
                              out_channels=self.out_channels,
                              stride=stride,
                              padding=padding,
                              bias=True
                              )
        self.add_module('_kernel', self._kernel)

        # Norm
        if self.norm:
            if self.norm == 'batch':
                self.add_module(
                    '_norm',
                    self.batch_norm(num_features=self.out_channels))
            elif self.norm == 'instance':
                self.add_module(
                    '_norm',
                    self.instance_norm(num_features=self.out_channels))

        # Activation/Non-Linearity
        if activation is not None:
            self.add_module('_activation', activation)
            if isinstance(activation, nn.SELU):
                self.weight_init = selu_weight_init_
                self.bias_init = selu_bias_init_
        # Pool
        if pool_size is not None:
            self.add_module(
                '_pool', self.pool_layer(kernel_size=pool_size))

        # Dropout
        if self.dropout is not None:
            if isinstance(activation, nn.SELU):
                self.add_module(
                    '_dropout', nn.AlphaDropout(self.dropout))
            else:
                self.add_module(
                    '_dropout', self.dropout_layer(self.dropout))
        self._init_weights()
        self._init_bias()

    def _init_layers(self):
        if self.conv_dim == 1:
            self.conv_layer = nn.Conv1d
            self.batch_norm = nn.BatchNorm1d
            self.instance_norm = nn.InstanceNorm1d
            self.pool_layer = nn.MaxPool1d
            self.dropout_layer = nn.Dropout
        elif self.conv_dim == 2:
            self.conv_layer = nn.Conv2d
            self.batch_norm = nn.BatchNorm2d
            self.instance_norm = nn.InstanceNorm2d
            self.pool_layer = nn.MaxPool2d
            self.dropout_layer = nn.Dropout2d
        elif self.conv_dim == 3:
            self.conv_layer = nn.Conv3d
            self.batch_norm = nn.BatchNorm3d
            self.instance_norm = nn.InstanceNorm3d
            self.pool_layer = nn.MaxPool3d
            self.dropout_layer = nn.Dropout3d
        else:
            self.conv_layer = None
            self.batch_norm = None
            self.instance_norm = None
            self.pool_layer = None
            self.dropout_layer = None
            raise ValueError(
                "Convolution is only supported for"
                " one of the first three dimensions.")

    def get_conv_output_size(self):
        """Calculate the size of the flattened features after conv."""
        with torch.no_grad():
            x = torch.ones(1, *self.in_dim)
            x = self.conv_model(x)
            return x.numel()


# TODO: Will work on these classes below later during Vulcan2 deployment
class InputUnit(BaseUnit):
    """InputUnit."""

    def __init__(self, in_channels, out_channels, bias=False):
        """Initialize InputUnit."""
        super(InputUnit, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self._kernel = nn.Linear(
            self.in_channels, self.out_channels, bias=bias)

    def forward(self, input):
        """Define forward for InputUnit."""
        if input.dim() > 2:
            input = input.transpose(1, 3)  # NCHW --> NHWC
            output = self._kernel(input)
            return output.transpose(1, 3)  # NHWC --> NCHW
        else:
            output = self._kernel(input)
            return output
